# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# yapf: disable
"""Module with pre-defined reusable commandline arguments that can be used as `click` decorators."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .main import *
from .overridable import *

__all__ = (
    'CALCULATION',
    'CALCULATIONS',
    'CODE',
    'CODES',
    'COMPUTER',
    'COMPUTERS',
    'CONFIG_OPTION',
    'DATA',
    'DATUM',
    'GROUP',
    'GROUPS',
    'INPUT_FILE',
    'LABEL',
    'NODE',
    'NODES',
    'OUTPUT_FILE',
    'OverridableArgument',
    'PROCESS',
    'PROCESSES',
    'PROFILE',
    'PROFILES',
    'USER',
    'WORKFLOW',
    'WORKFLOWS',
)

# yapf: enable
