# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Testing infrastructure for easy testing of AiiDA plugins.
"""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .main import *

__all__ = (
    'ProfileManager',
    'TemporaryProfileManager',
    'TestManager',
    'TestManagerError',
    'get_test_backend_name',
    'get_test_profile_name',
    'get_user_dict',
    'test_manager',
)

# yapf: enable
